/*	File:	  cmix.h                (this is -*- c++ -*- code)
 *      Author:	  Lars Ole Andersen (lars@diku.dk)
 *                Peter Holst Andersen (txix@diku.dk)
 *                Henning Makholm (makholm@diku.dk)
 *      Content:  C-Mix Cogen library.
 *
 *      Copyright  1998. The TOPPS group at DIKU, U of Copenhagen.
 *      Redistribution and modification are allowed under certain
 *      terms; see the file COPYING.cmix for details.
 */

#ifndef __cmix__
#define __cmix__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
/*	File:	  rusage.h
 *      Author:	  Peter Holst Andersen (txix@diku.dk)
 *      Date:	  February 1996
 *      Content:  C-Mix Cogen library: Resource usage
 *
 *      Copyright  1998. The TOPPS group at DIKU, U of Copenhagen.
 *      Redistribution and modification are allowed under certain
 *      terms; see the file COPYING.cmix for details.
 */

#ifndef __rusage__
#define __rusage__
  
void check_memusage(void);
void print_memusage(void);

void collect_statistic(int, int =0);

#define STAT_CMP_STATE		0
#define STAT_CMP_STATE_MATCH	1
#define STAT_LOOKUP_SHARED	2
#define STAT_LOOKUP_SHARED2	3
#define STAT_SHARED_FOUND	4
#define STAT_ALLOCATED		5
#define STAT_CMP_NULL		6
#define STAT_CMP_CHECK_GLOBALS  7
#define STAT_CMP_GLOBALS	8
#define STAT_CMP_CHECK_STACK	9
#define STAT_CMP_STACK		10
#define STAT_CMP_CHECK_HEAP	11
#define STAT_CMP_HEAP		12
#define STAT_CMP_NOT_FOUND	13
#define STAT_COPY_NULL		14
#define STAT_COPY_CHECK_GLOBALS	15
#define STAT_COPY_GLOBALS	16
#define STAT_COPY_CHECK_STACK	17
#define STAT_COPY_STACK		18
#define STAT_COPY_CHECK_HEAP	19
#define STAT_COPY_HEAP		20
#define STAT_COPY_NOT_FOUND	21
#define STAT_LAST		22

#endif /* __rusage__ */
/*	File:	  iu.h                (this is -*- c++ -*- code)
 *      Author:	  Arne John Glenstrup
 *      Created:  03.06.1998
 *      Content:  C-Mix Cogen library: in-use information manipulation
 *
 *      Copyright  1998. The TOPPS group at DIKU, U of Copenhagen.
 *      Redistribution and modification are allowed under certain
 *      terms; see the file COPYING.cmix for details.
 */

#ifndef __IU_H__
#define __IU_H__

typedef char* IUInfo;

#define notInUse(ID, inUse) (inUse && inUse[ID] == '0')


#endif /* __IU_H__ */
/*	File:	  ro.h                (this is -*- c++ -*- code)
 *      Author:	  Arne John Glenstrup
 *      Created:  03.06.1998
 *      Content:  C-Mix Cogen library: in-use information manipulation
 *
 *      Copyright  1998. The TOPPS group at DIKU, U of Copenhagen.
 *      Redistribution and modification are allowed under certain
 *      terms; see the file COPYING.cmix for details.
 */

#ifndef __RO_H__
#define __RO_H__

typedef char* ROInfo;

extern unsigned int cmixMWCount[];

#define isReadOnly(ID, mayWrite) (mayWrite && mayWrite[ID] == '0' && cmixMWCount[ID] == 0)


#endif /* __RO_H__ */

#define cmixSPECLIB_VERSION 2003

struct cmixMemoWhat {
    IUInfo IU ;
    ROInfo RO ;
} ;

//---------------------------------------------------------------------------
// Memory management
//---------------------------------------------------------------------------

class cmixDataObject;

void cmixPushObject(cmixDataObject *);
void cmixPopObjects(int);
void cmixRegisterHeapObject(cmixDataObject *);

struct cmixDataObjectCopy {
    void *copy;				// Copied data
    size_t size;			// Size of the copied object
    cmixDataObject *obj;		// The original object desc.
    long offset ;			// offset always memoized
    cmixDataObjectCopy *next;		// Next element in the list

    cmixDataObjectCopy();
    cmixDataObjectCopy(cmixDataObject *,void const *);
};

class cmixPointerCallback {
    typedef bool (*func)(cmixDataObjectCopy *&,void const*,cmixMemoWhat&);
    friend class cmixDataObject ;
    cmixDataObjectCopy *&copy;
    cmixMemoWhat &what;
    const func F ;
    cmixPointerCallback(cmixDataObjectCopy *&,cmixMemoWhat &,func f);
public:
    inline bool operator()(void const*d) { return F(copy,d,what); }
};

typedef bool (*cmixPointerFollower)(cmixPointerCallback &,void*,unsigned);

class cmixDataObject {
  bool maycopy(cmixMemoWhat&);
  cmixPointerFollower follow_pointers ;
  const unsigned array_length ;
public:
  void *obj;			// The data object
  const unsigned long ID;       // Its unique ID
  size_t size;			// The size of the data object
  bool visit;			// Visit flag
   
  cmixDataObjectCopy *copy_me(void const*, cmixMemoWhat&);
  bool compare_me(cmixDataObjectCopy *&, void const*, cmixMemoWhat&);
    
  cmixDataObject(unsigned long,void*,size_t,
                 cmixPointerFollower,unsigned,
                 bool push=true);
    
  ~cmixDataObject();
};

struct cmixState {
    cmixDataObject **globals, **locals;	// Null terminated arrays of objects

    cmixState(cmixDataObject **g, cmixDataObject **l)
    {
	globals = g;
	locals = l;
    }
    ~cmixState(void) { globals = NULL; locals = NULL; }
};

struct cmixStateCopy {
    cmixDataObjectCopy *objs;		// List of copies of objects

    cmixStateCopy();
    ~cmixStateCopy();			// Delete self and the objs list
};

// ========================= Copy, Restore, Compare =========================

void cmixEndOfGlobals(void);   // Indicates that all globals have been declared
cmixStateCopy *cmixCopyState(cmixState *, cmixMemoWhat&);
cmixStateCopy *cmixCopyLightState(cmixState *, cmixMemoWhat&);
cmixStateCopy *cmixExtractEndState(cmixStateCopy *,cmixState *);
void cmixRestoreState(cmixStateCopy *);		  // Restore saved state
bool cmixCmpState(cmixState *, cmixStateCopy *, cmixMemoWhat&);
bool cmixCmpLightState(cmixStateCopy*);

// Basic pointer-following function
bool cmix_follow1(cmixPointerCallback &,void*,unsigned);

//---------------------------------------------------------------------------
// Code type
//---------------------------------------------------------------------------

union cmixExpr ;

typedef union cmixExpr* Code ;

inline void cmix_putName(Code &cmixThis,Code cmixIt) { cmixThis = cmixIt; }
Code cmixMkExp(const char*,...);

// Lift operators
Code cmixLift(char);
Code cmixLift(unsigned char);
Code cmixLift(signed char);
Code cmixLift(short);
Code cmixLift(unsigned short);
Code cmixLift(int);
Code cmixLift(unsigned int);
Code cmixLift(long);
Code cmixLift(unsigned long);
Code cmixLift(float);
Code cmixLift(double);
Code cmixLift(long double);
Code cmixLift(char const*); // Lift string - not used presently
Code cmixLift(Code *); // Lift pointer to dynamic data

//---------------------------------------------------------------------------
// Code generation
//---------------------------------------------------------------------------

// Declarations
void cmixGlobal(Code,const char *,...);
void cmixParam(Code,const char *,...);
void cmixLocal(Code,const char *,...);
void cmixMemberDecl(Code,const char*,...);

// cmixIntro marks declarations that should ALWAYS be output.
extern Code const cmixIntro ;

// Residual statement generators (they emit code to the current function)
typedef union cmixStmt *cmixLabelTag ;

cmixLabelTag cmixMakeLabel() ;
void cmixLabel(cmixLabelTag);
void cmixStmt(const char *,...); // expression stmt - speclib supplies the ";"
void cmixGoto(cmixLabelTag);
void cmixIf(cmixLabelTag thenLabel, cmixLabelTag elseLabel, const char *,...);
void cmixReturn();	           	// Generate return without expr
void cmixReturn(const char *,...);	// Generate return with expr

// Functions
enum cmixFcnMemoMode { cmixSharedPure,
                       cmixSharedSideEffecting,
                       cmixNonshared };
Code cmixFunctionSeenB4(char *, cmixState *, IUInfo, ROInfo);
void cmixPushFun(const char *, cmixState *, IUInfo, ROInfo,
                  Code residname, cmixFcnMemoMode, const char *type,...);
void cmixPopFun(cmixState *,cmixFcnMemoMode);


void *cmixAllocStatic
   (unsigned long ID, size_t, unsigned nelem, cmixPointerFollower);

//---------------------------------------------------------------------------
// Pending
//---------------------------------------------------------------------------

cmixLabelTag cmixPendinsert(void *, cmixState*, IUInfo, ROInfo);
					// Insert into pending list
void *cmixPending(void);		// Check pending list
void cmixPushPend(void);		// Push current lists onto stack
void cmixPopPend(void);			// Pop from stack

//---------------------------------------------------------------------------
// Unparsing, name management
//---------------------------------------------------------------------------

struct cmixNameRec {
    char const *base ;
    unsigned nextseq_global ;
    unsigned nextseq_local ;
} ;
Code cmixRequestName(unsigned index,unsigned wseq);
void cmixStructMemA(Code,struct cmixNameRec[]);
void cmixStructMemB(Code,struct cmixNameRec[]);

void cmixUnparse(cmixNameRec names[],unsigned const namesSorted[],
                 unsigned cNames, FILE *);

//---------------------------------------------------------------------------
// Errors etc
//---------------------------------------------------------------------------
void cmixFatal(char *);		// Print error mesg. and abort.

//---------------------------------------------------------------------------
// Utility function to convert a one length string to a char (called by main
// function in generating extension)
//---------------------------------------------------------------------------
char cmix_atoc(char *);
long cmix_atoi(char *);
double cmix_atof(char *);

#define cmixCONCAT2(a,b) cmixCONCAT(a,b)
#define cmixCONCAT(a,b) a ## b
#define cmixSpeclibInit cmixCONCAT2(cmixVersionStamp,cmixSPECLIB_VERSION)
void cmixSpeclibInit();

extern "C" void cmixGenExit(FILE*) ; // may be forward referenced

#endif /* __cmix__ */
